/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.xslt;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.mail.util.ByteArrayDataSource;
import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.lang.CharEncoding;
import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.xslt.model.XsltConfigurationHandler;

/**
 * The JBI Listener for the XSLT service engine.
 *
 * @author Olivier Fabre - EBM WebSourcing
 * @author Adrien Louis - EBM WebSourcing
 * @author cgirodengo
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class XsltJBIListener extends AbstractJBIListener {

    /**
     * The (local) name of the 'transform' operation.
     */
    private static final String OPERATION_TRANSFORM = "transform";

    /**
     * The (local) name of the 'transformToAttachment' operation.
     * <p>
     * This method is kept to be coherent with previous versions (and thus avoid an update of existing SU).
     * But this operation is not exposed in the WSDL of SU for this component.
     * </p>
     */
    // private static final String OPERATION_TRANSFORM_TO_ATTACHMENT = "transformToAttachment";

    /**
     * The (local) name of the 'transformToMtomAttachment' operation.
     * <p>
     * This method is the one to use to attach the transformation result to the response.
     * </p>
     */
    private static final String OPERATION_TRANSFORM_TO_MTOM_ATTACHMENT = "transformToMtomAttachment";

    /**
     * The name space for XOP attachments.
     */
    public static final String MTOM_TNS = "http://www.w3.org/2004/08/xop/include";

    /**
     * A prefix to use for logged messages.
     * <p>
     * This prefix is updated for every processed message.<br />
     * It only contains the exchange ID.
     * </p>
     */
    private String logHint;



    /*
     * (non-Javadoc)
     * @see org.ow2.petals.component.framework.listener.AbstractJBIListener
     * #onJBIMessage(org.ow2.petals.component.framework.api.message.Exchange)
     */
    @Override
    public boolean onJBIMessage(final Exchange exchange) {

        Logger log = getLogger();
        this.logHint = "Exchange " + exchange.getExchangeId();

        if( log.isLoggable( Level.INFO ))
            log.info( this.logHint + " was received and is started to be processed." );

        Exception finalException = null;
        try {
            // Is the exchange valid?
            String errorMsg;
            if ((errorMsg = validateExchange(exchange)) != null) {
                setFault(exchange, errorMsg, null);
            }

            // Valid exchange => process it
            else {
                try {
                    this.processExchange( exchange );
                    if( log.isLoggable( Level.INFO ))
                        log.info( this.logHint + " was successfully processed." );

                } catch (final IOException e) {
                    setFault(exchange, null, e);

                } catch (MessagingException e) {
                    setFault(exchange, null, e);

                } catch (TransformerException e) {
                    setFault(exchange, null, e);
                }
            }

        } catch (PEtALSCDKException e) {
            finalException = new MessagingException(ExceptionUtil.getExtendedMessage(e));

        } catch (MessagingException e) {
            finalException = e;
        }

        // Handle cases where a fault could not be set on the exchange
        if (finalException != null) {
            if (log.isLoggable(Level.INFO))
                log.info(this.logHint + " encountered a problem. " + finalException.getMessage());

            // Technical error, it would be set as a Fault by the CDK
            if (exchange.isActiveStatus()) {
                exchange.setError(finalException);
            }
        }

        return true;
    }

    /**
     * Validates the operation and the MEP.
     * <p>
     * No logs is written by this method. It is the responsibility of the
     * callers to log messages.
     * </p>
     *
     * @param exchange
     *            the exchange to validate
     * @return null if the exchange is valid, an error message otherwise
     */
    private static final String validateExchange(final Exchange exchange) {

        String errorMsg = null;
        final QName operation = exchange.getOperation();

        // Null operation
        if( operation == null ) {
            errorMsg = "The JBI operation must be set by the consumer of the service.";
        }

        // Transform operation
        else if( OPERATION_TRANSFORM.equalsIgnoreCase( operation.getLocalPart())) {
            if( ! Message.MEPConstants.IN_OUT_PATTERN.equals( exchange.getPattern())) {
                errorMsg = "The operation '" + operation + "' must use the pattern 'InOut'.";
            }
        }

        // Transform to attachment operation: depreciated
        // It is commented in case where a complaint would be posted
        //		else if( OPERATION_TRANSFORM_TO_ATTACHMENT.equalsIgnoreCase( operation.getLocalPart())) {
        //			if( ! Message.MEPConstants.IN_OUT_PATTERN.equals( exchange.getPattern())) {
        //				String msg = "The operation '" + operation + "' must use the pattern 'InOut'.";
        //				exchange.setError(new Exception( msg ));
        //				result = false;
        //			}
        //		}

        // Transform to MTOM attachment operation
        else if( OPERATION_TRANSFORM_TO_MTOM_ATTACHMENT.equalsIgnoreCase( operation.getLocalPart())) {
            if( ! Message.MEPConstants.IN_OUT_PATTERN.equals( exchange.getPattern())) {
                errorMsg = "The operation '" + operation + "' must use the pattern 'InOut'.";
            }
        }

        // Unsupported operation
        else {
            errorMsg = "The operation '" + operation + "' is not supported by this component.";
        }

        return errorMsg;
    }


    /**
     * Processes the transformation and sets the response as Source or attachment.
     *
     * @param exchange the exchange to process
     * @throws MessagingException if the exchange properties could not be read
     * @throws TransformerException if the XML transformation failed
     * @throws IOException if the transformation result could not be returned as an attachment
     */
    private void processExchange( final Exchange exchange )
    throws MessagingException, TransformerException, IOException {

        Logger log = getLogger();
        if( log.isLoggable( Level.INFO ))
            log.info( this.logHint + ": the request was validated, the transformation is about to start." );


        // Get the configuration handler
        XsltConfigurationHandler handler = ((XsltSe) this.component).getXsltConfigurationHandler( exchange.getEndpointName());
        if( handler == null ) {
            String msg = this.logHint + " encountered a problem. No handler was found for this end-point.";
            throw new TransformerException( msg );
        }


        // Get the input message content
        final Source inputSource = exchange.getInMessageContentAsSource();
        if( log.isLoggable( Level.FINEST )) {
            try {
                // The transformation result is shown by the CDK when the response is sent (debug level)
                String inputMsg = UtilFactory.getSourceUtil().createString( inputSource, true );
                log.finest( this.logHint + ": the message to transform is\n" + inputMsg );

            } catch( PEtALSCDKException e ) {
                log.finest( this.logHint + ": the message to transform could not be displayed, an error occurred." );
            }
        }

        // Perform the transformation
        ClassLoader customClassLoader = handler.getClassLoader();
        Transformer transformer = handler.createTransformer(exchange, log);
        if( customClassLoader != null )
            Thread.currentThread().setContextClassLoader( customClassLoader );

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        Result outputResult = new StreamResult(outputStream);
        try {
            transformer.transform(inputSource, outputResult);

        } finally {
            if (customClassLoader != null)
                Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        }

        // How do we return the transformation result?
        String opName = exchange.getOperation().getLocalPart();

        // The result is in the message's content
        if( OPERATION_TRANSFORM.equalsIgnoreCase( opName )) {

            // Keep the machine's default encoding
            exchange.setOutMessageContent( new ByteArrayInputStream( outputStream.toByteArray()));
        }

        // The result is returned as an attachment (not MTOM)
        // This method is kept to be coherent with previous versions (and thus avoid an update of existing SU).
        // But this operation is not exposed in the WSDL of SU for this component.
        // Depreciated
        // It is commented in case where a complaint would be posted
        //		else if( OPERATION_TRANSFORM_TO_ATTACHMENT.equalsIgnoreCase( opName )){
        //			ByteArrayInputStream input = new ByteArrayInputStream( outputStream.toByteArray());
        //			DataSource datasource = new ByteArrayDataSource( input, "application/octet-stream" );
        //			DataHandler attachmentHandler = new DataHandler( datasource );
        //
        //			String attachmentName = handler.getOutputAttachmentName();
        //			exchange.setOutMessageAttachment( attachmentName, attachmentHandler );
        //
        //			String outSource =
        //				"<?xml version=\"1.0\" encoding=\"UTF-8\"?><attached-files><file-name>"
        //				+ attachmentName + "</file-name></attached-files>";
        //
        //			// Keep the machine's default encoding
        //			exchange.setOutMessageContent( new ByteArrayInputStream( outSource.getBytes( "UTF-8" )));
        //		}

        // The result is returned as an attachment (MTOM)
        else if( OPERATION_TRANSFORM_TO_MTOM_ATTACHMENT.equalsIgnoreCase( opName )){
            ByteArrayInputStream input = new ByteArrayInputStream( outputStream.toByteArray());
            DataSource datasource = new ByteArrayDataSource( input, "application/octet-stream" );
            DataHandler attachmentHandler = new DataHandler( datasource );

            String attachmentName = handler.getOutputAttachmentName();
            exchange.setOutMessageAttachment( attachmentName, attachmentHandler );

            StringBuilder sb = new StringBuilder();
            sb.append( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" );
            sb.append( "<attachedTransformResponse " );
            sb.append( "xmlns=\"" + exchange.getOperation().getNamespaceURI() + "\" " );
            sb.append( "xmlns:xop=\"" + MTOM_TNS + "\">\n" );
            sb.append( "\t<fileContent>\n" );
            sb.append( "\t\t<xop:include href=\"cid:" + attachmentName + "\" />\n" );
            sb.append( "\t</fileContent>\n" );
            sb.append( "</attachedTransformResponse>" );

            // Keep the machine's default encoding
            exchange.setOutMessageContent( new ByteArrayInputStream( sb.toString().getBytes( "UTF-8" )));
        }
    }

    /**
     * Builds a fault and sets it on the exchange.
     *
     * @param exchange
     *            the message exchange
     * @param faultMessage
     *            the fault message
     * @param e
     *            an exception which caused this fault
     * @throws PEtALSCDKException
     *             if a source could not be created
     * @throws MessagingException
     *             if the fault could not be set on the exchange
     */
    private void setFault(Exchange exchange, String faultMessage, Throwable e) throws PEtALSCDKException,
            MessagingException {

        // Build an error message
        final String errorMsg;
        if (faultMessage != null)
            errorMsg = faultMessage;
        else if (e != null && e.getMessage() != null)
            errorMsg = e.getMessage();
        else
            errorMsg = "An error occurred in the XSLT SE.";

        // Build the fault document
        StringBuilder sb = new StringBuilder();
        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<xsltFault ");
        sb.append("xmlns=\"" + exchange.getOperation().getNamespaceURI() + "\">\n");
        sb.append("\t<message>" + errorMsg + "</message>\n");
        if (e != null) {
            StringWriter sWriter = new StringWriter();
            e.printStackTrace(new PrintWriter(sWriter));
            sb.append("\t<details>\n" + sWriter.getBuffer().toString() + "\n\t</details>\n");
        }
        sb.append("</xsltFault>");

        // Create the stream source
        Source source = UtilFactory.getSourceUtil().createSource(sb.toString(), CharEncoding.UTF_8);

        // Create and set the fault
        Fault fault = exchange.createFault();
        fault.setContent(source);
        exchange.setFault(fault);
    }
}
